(*
 * Builtin operations on files.
 *
 * \begin{doc}
 * \section{File operations}
 * \end{doc}
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_glob
open Lm_printf
open Lm_location

open Omake_ir
open Omake_env
open Omake_eval
open Omake_node
open Omake_rule
open Omake_value
open Omake_target
open Omake_symbol
open Omake_build_util
open Omake_builtin_util

module Pos = MakePos (struct let name = "Omake_builtin_file" end)
open Pos

(*
 * Get the file from a string.
 *
 * \begin{doc}
 * \subsection{file, dir}
 *
 * \begin{verbatim}
 *    $(file sequence) : File Sequence
 *       sequence : Sequence
 *    $(dir sequence) : Dir Sequence
 *       sequence : Sequence
 * \end{verbatim}
 *
 * The \verb+file+ and \verb+dir+ functions define location-independent references to files and directories.
 * In \Prog{omake}, the commands to build a target are executed in the target's directory.  Since there may be
 * many directories in an \Prog{omake} project, the build system provdes a way to construct a reference to a file
 * in one directory, and use it in another without explicitly modifying the file name.  The functions have the following
 * syntax, where the name should refer to a file or directory.
 *
 * For example, we can construct a reference to a file \verb+foo+ in the current directory.
 *
 * \begin{verbatim}
 *    FOO = $(file foo)
 *    .SUBDIRS: bar
 * \end{verbatim}
 *
 * If the \verb+FOO+ variable is expanded in the \verb+bar+ subdirectory, it will expand to \verb+../foo+.
 *
 * These commands are often used in the top-level OMakefile to provide location-independent references to
 * top-level directories, so that build commands may refer to these directories as if they were absolute.
 *
 * \begin{verbatim}
 *    ROOT = $(dir .)
 *    LIB  = $(dir lib)
 *    BIN  = $(dir bin)
 * \end{verbatim}
 *
 * Once these variables are defined, they can be used in build commands in subdirectories as follows, where
 * \verb+$(BIN)+ will expand to the location of the \verb+bin+ directory relative to the command being executed.
 *
 * \begin{verbatim}
 *    install: hello
 * 	cp hello $(BIN)
 * \end{verbatim}
 * \end{doc}
 *)
let file venv pos loc args =
   let pos = string_pos "file" pos in
      match args with
         [arg] ->
            let values = values_of_value venv pos arg in
            let values = List.map (node_value_of_value venv pos) values in
               concat_array values
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

let dir venv pos loc args =
   let pos = string_pos "dir" pos in
      match args with
         [arg] ->
            let values = values_of_value venv pos arg in
            let values = List.map (node_value_of_value venv pos) values in
               concat_array values
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * Display something from a different directory.
 *
 * \begin{doc}
 * \subsection{in}
 *
 * \begin{verbatim}
 *    $(in dir, exp) : String Array
 *       dir : Dir
 *       exp : expression
 * \end{verbatim}
 *
 * The \verb+in+ function is closely related to the \verb+dir+ and
 * \verb+file+ functions.  It takes a directory and an expression, and
 * evaluates the expression in that effective directory.
 * For example, one
 * common way to install a file is to define a symbol link, where the
 * value of the link is relative to the directory where the link is
 * created.
 *
 * The following commands create links in the \verb+$(LIB)+ directory.
 *
 * \begin{verbatim}
 *     FOO = $(file foo)
 *     install:
 *        ln -s $(in $(LIB), $(FOO)) $(LIB)/foo
 * \end{verbatim}
 * \end{doc}
 *)
let ind venv pos loc args =
   let pos = string_pos "ind" pos in
      match args with
         [dir; arg] ->
            let dir = dir_of_value venv pos dir in
            let venv = venv_chdir_dir venv dir in
            let strings = strings_of_value venv pos arg in
               concat_strings strings
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))

(*
 * Search the PATH.
 *
 * \begin{doc}
 * \subsection{which}
 *
 * \begin{verbatim}
 *    $(which files) : File Sequence
 *       files : String Sequence
 * \end{verbatim}
 *
 * The \verb+which+ function searches for executables in the
 * current command search path, and returns \verb+file+ values
 * for each of the commands.  It is an error if a command is
 * not found.
 * \end{doc}
 *)
let which venv pos loc args =
   let pos = string_pos "which" pos in
      match args with
         [arg] ->
            let args = strings_of_value venv pos arg in
            let args =
               List.map (fun s ->
                     try Lm_filename_util.which s with
                        Not_found ->
                           raise (OmakeException (loc_pos loc pos, StringStringError ("command not found", s)))) args
            in
            let args = List.map (fun s -> ValNode (venv_intern venv s)) args in
               concat_array args
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * \begin{doc}
 * \subsection{exists-in-path}
 *
 * \begin{verbatim}
 *    $(exists-in-path files) : String
 *       files : String Sequence
 * \end{verbatim}
 *
 * The \verb+exists-in-path+ function tests whether all executables
 * are present in the current search path.
 * \end{doc}
 *)
let exists_in_path venv pos loc args =
   let pos = string_pos "exists-in-path" pos in
      match args with
         [arg] ->
            let args = strings_of_value venv pos arg in
            let test =
               List.for_all (fun s ->
                     try ignore (Lm_filename_util.which s); true with
                        Failure _
                      | Not_found ->
                           false) args
            in
               if test then
                  val_true
               else
                  val_false
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * Strip the directory.
 *
 * \begin{doc}
 * \subsection{basename}
 *
 * \begin{verbatim}
 *    $(basename files) : String Sequence
 *       files : String Sequence
 * \end{verbatim}
 *
 * The \verb+basename+ function returns the base names for a list of files.
 * The basename is the filename without any leading directory components removed.
 *
 * For example, the expression \verb+$(basename dir1/dir2/a.out /etc/modules.conf /foo.ml)+ evaluates to
 * \verb+a.out modules.conf foo.ml+.
 * \end{doc}
 *)
let basename venv pos loc args =
   let pos = string_pos "basename" pos in
      match args with
         [arg] ->
            let args = strings_of_value venv pos arg in
            let args = List.map Filename.basename args in
               concat_strings args
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * Get the directory.
 *
 * \begin{doc}
 * \subsection{dirof}
 *
 * \begin{verbatim}
 *    $(dirof files) : Dir Sequence
 *       files : File Sequence
 * \end{verbatim}
 *
 * The \verb+dirof+ function returns the directory for each of the listed files.
 *
 * For example, the expression \verb+$(dirof dir/dir2/a.out /etc/modules.conf /foo.ml)+ evaluates
 * to the directories \verb+dir1/dir2 /etc /+.
 * \end{doc}
 *)
let dirof venv pos loc args =
   let pos = string_pos "dirof" pos in
      match args with
         [arg] ->
            let values = values_of_value venv pos arg in
            let nodes = List.map (file_of_value venv pos) values in
            let dirs = List.map (fun node -> ValDir (Node.dir node)) nodes in
               concat_array dirs
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * Strip the directory.
 *
 * \begin{doc}
 * \subsection{rootname}
 *
 * \begin{verbatim}
 *    $(rootname files) : String Sequence
 *       files : String Sequence
 * \end{verbatim}
 *
 * The \verb+rootname+ function returns the root name for a list of files.
 * The rootname is the filename with the final suffix removed.
 *
 * For example, the expression \verb+$(rootname dir1/dir2/a.out /etc/a.b.c /foo.ml)+ evaluates to
 * \verb+dir1/dir2/a /etc/a.b /foo+.
 * \end{doc}
 *)
let rootname venv pos loc args =
   let pos = string_pos "rootname" pos in
      match args with
         [arg] ->
            let args = strings_of_value venv pos arg in
            let args = List.map Lm_filename_util.root args in
               concat_strings args
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * Strip the directory.
 *
 * \begin{doc}
 * \subsection{homename}
 *
 * \begin{verbatim}
 *    $(homename files) : String Sequence
 *       files : File Sequence
 * \end{verbatim}
 *
 * The \verb+homename+ function returns the name of a file in
 * tilde form, if possible.  The unexpanded forms are computed
 * lazily: the \verb+homename+ function will usually evaluate to an absolute
 * pathname until the first tilde-expansion for the same directory.
 * \end{doc}
 *)
let homename venv pos loc args =
   let pos = string_pos "rootname" pos in
      match args with
         [arg] ->
            let args = values_of_value venv pos arg in
            let venv = venv_chdir_dir venv Dir.root in
            let args =
               List.map (fun v ->
                     ValString (tilde_collapse (string_of_value venv pos v))) args
            in
               concat_array args
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * Strip the directory.
 *
 * \begin{doc}
 * \subsection{suffix}
 *
 * \begin{verbatim}
 *    $(suffix files) : String Sequence
 *       files : StringSequence
 * \end{verbatim}
 *
 * The \verb+suffix+ function returns the suffixes for a list of files.
 * If a file has no suffix, the function returns the empty string.
 *
 * For example, the expression \verb+$(suffix dir1/dir2/a.out /etc/a /foo.ml)+ evaluates
 * to \verb+.out .ml+.
 * \end{doc}
 *)
let suffix venv pos loc args =
   let pos = string_pos "suffix" pos in
      match args with
         [arg] ->
            let args = strings_of_value venv pos arg in
            let args = List.map Lm_filename_util.suffix args in
               concat_strings args
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * Check if a file exists.
 *
 * \begin{doc}
 * \subsection{file-exists}
 *
 * \begin{verbatim}
 *    $(file-exists files) : String
 *       files : File Sequence
 * \end{verbatim}
 *
 * The \verb+file-exists+ function checks whether the files listed exist.
 *
 * \subsection{target-exists}
 *
 * \begin{verbatim}
 *    $(target-exists files) : String
 *       files : File Sequence
 * \end{verbatim}
 *
 * The \verb+target-exists+ function is similar to the \verb+file-exists+ function.
 * However, it returns true if the file exists \emph{or} if it can be built
 * by the current project.
 *
 * \end{doc}
 *)
let node_exists node_exists venv pos loc args =
   let pos = string_pos "file-exists" pos in
      match args with
         [arg] ->
            let cache = venv_cache venv in
            let args = strings_of_value venv pos arg in
            let b =
               List.for_all (fun arg ->
                     node_exists cache venv (venv_intern venv arg)) args
            in
               if b then
                  val_true
               else
                  val_false
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * Filter out the files that don't exist.
 *
 * \begin{doc}
 * \subsection{filter-exists, filter-targets}
 *
 * \begin{verbatim}
 *    $(filter-exists files) : File Sequence
 *       files : File Sequence
 *    $(target-exists files) : File Sequence
 *       files : File Sequence
 * \end{verbatim}
 *
 * The \verb+filter-exists+ and \verb+filter-targets+ functions remove files from
 * a list of files if they do not exist (\verb+filter-exists+) or they do not
 * exist and can't be built according to the current project (\verb+filter-targets+).
 * \end{doc}
 *)
let filter_nodes node_exists venv pos loc args =
   let pos = string_pos "filter-exists" pos in
      match args with
         [arg] ->
            let cache = venv_cache venv in
            let args  = values_of_value venv pos arg in
            let nodes = List.map (file_of_value venv pos) args in
            let nodes = List.filter (node_exists cache venv) nodes in
            let nodes = List.map (fun v -> ValNode v) nodes in
               ValArray nodes
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

let file_exists venv pos loc args =
   node_exists (fun cache venv node -> Omake_cache.exists cache node) venv pos loc args

let filter_exists venv pos loc args =
   filter_nodes (fun cache venv node -> Omake_cache.exists cache node) venv pos loc args

let target_exists venv pos loc args =
   node_exists target_is_buildable venv pos loc args

let filter_targets venv pos loc args =
   filter_nodes target_is_buildable venv pos loc args

(*
 * Get the file from a string.
 *
 * \begin{doc}
 * \subsection{file-sort}
 *
 * \begin{verbatim}
 *    $(file-sort order, files) : File Sequence
 *       order : String
 *       files : File Sequence
 * \end{verbatim}
 *
 * The \verb+file-sort+ function sorts a list of filenames by
 * build order augmented by a set of sort rules.  Sort
 * rules are declared using the \verb+.ORDER+ target.
 * The \verb+.BUILDORDER+ defines the default order.
 *
 * \verb+$(file-sort <order>, <files>)+
 *
 * For example, suppose we have the following set of rules.
 *
 * \begin{verbatim}
 *    a: b c
 *    b: d
 *    c: d
 *
 *    .DEFAULT: a b c d
 *       echo $(file-sort .BUILDORDER, a b c d)
 * \end{verbatim}
 *
 * In the case, the sorter produces the result \verb+d b c a+.
 * That is, a target is sorted \emph{emph} after its dependencies.
 * The sorter is frequently used to sort files that are to be linked
 * by their dependencies (for languages where this matters).
 *
 * There are three important restrictions to the sorter:
 * \begin{itemize}
 * \item The sorter can be used only within a rule body.
 * The reason for this is that \emph{all} dependencies
 * must be known before the sort is performed.
 * \item The sorter can only sort files that are buildable
 * in the current project.
 * \item The sorter will fail if the dependencies are cyclic.
 * \end{itemize}
 *
 * \subsection{sort rule}
 *
 * It is possible to firther constrain the sorter through the use of
 * sort rules.  A sort rule is declared in two steps.  The
 * target must be listed as an \verb+.ORDER+ target; and then
 * a set of sort rules must be given.  A sort rule defines
 * a pattern constraint.
 *
 * \begin{verbatim}
 *    .ORDER: .MYORDER
 *
 *    .MYORDER: %.foo: %.bar
 *    .MYORDER: %.bar: %.baz
 *
 *    .DEFAULT: a.foo b.bar c.baz d.baz
 *       echo $(sort .MYORDER, a.foo b.bar c.baz d.baz)
 * \end{verbatim}
 *
 * In this example, the \verb+.MYORDER+ sort rule specifies that any
 * file with a suffix \verb+.foo+ should be placed after any file with
 * suffix \verb+.bar+, and any file with suffix \verb+.bar+ should be
 * placed after a file with suffix \verb+.baz+.
 *
 * In this example, the result of the sort is \verb+d.baz c.baz b.bar a.foo+.
 * \end{doc}
 *)
let sort_aux sorter venv pos loc args =
   let pos = string_pos "file-sort" pos in
   let name, nodes =
      match args with
         [name; arg] ->
            let values = values_of_value venv pos arg in
            let nodes = List.map (file_of_value venv pos) values in
            let name = Lm_symbol.add (string_of_value venv pos name) in
               name, nodes
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))
   in
   let env = get_env pos loc in
      sorter env venv pos name nodes

let sort venv pos loc args =
   let sorter env venv pos name nodes =
      let nodes = Omake_build_util.sort env venv pos name nodes in
         ValSequence (sequence_map (fun node -> ValNode node) nodes)
   in
      sort_aux sorter venv pos loc args

(*
 * \begin{doc}
 * \subsection{file-check-sort}
 *
 * \begin{verbatim}
 *    file-check-sort(files)
 *       files : File Sequence
 *    raises RuntimeException
 * \end{verbatim}
 *
 * The \verb+file-check-sort+ function checks whether a list of files
 * is in sort order.  If so, the list is returned unchanged.
 * If not, the function raises an exception.
 *
 * \verb+$(file-check-sort <order>, <files>)+
 * \end{doc}
 *)
let check_sort venv pos loc args =
   let sorter env venv pos name nodes =
      Omake_build_util.check_sort env venv pos name nodes;
      ValSequence (sequence_map (fun node -> ValNode node) nodes)
   in
      sort_aux sorter venv pos loc args

(************************************************************************
 * Listings.
 *)

(*
 * Comparisons for forting.
 *)
let compare_dir_node dir1 node =
   let dir2 = Node.dir node in
   let cmp = DirCompare.compare_alpha dir1 dir2 in
      if cmp = 0 then
         -1
      else
         cmp

let compare_val_nodes node1 node2 =
   match node1, node2 with
      ValDir dir, ValNode node ->
         compare_dir_node dir node
    | ValNode node, ValDir dir ->
         -(compare_dir_node dir node)
    | ValDir dir1, ValDir dir2 ->
         DirCompare.compare_alpha dir1 dir2
    | ValNode node1, ValNode node2 ->
         NodeCompare.compare_alpha node1 node2
    | _ ->
         0

let sort_val_nodes nodes =
   List.sort compare_val_nodes nodes

(*
 * \begin{doc}
 * \subsection{glob}
 *
 * \begin{verbatim}
 *    $(glob strings) : Node Array
 *       strings : String Sequence
 *    $(glob options, strings) : Node Array
 *       options : String
 *       strings : String Sequence
 * \end{verbatim}
 *
 * The \verb+glob+ function performs glob-expansion.
 *
 * The . and .. entries are always ignored.
 *
 * The options are:
 * \begin{description}
 * \item[b] Do not perform \Cmd{csh}{1}-style brace expansion.
 * \item[c] If a glob expansion does not match, return the pattern instead of aborting.
 * \item[.] Allow wildcard patterns to match files beginning with a .
 * \item[A] Return all files, including files that begin with a .
 * \item[D] Match only directory files.
 * \item[C] Ignore files according to \Cmd{cvs}{1} rules.
 * \end{description}
 *
 * In addition, the following variables may be defined that affect the
 * behavior of \verb+glob+.
 *
 * \begin{description}
 * \item[GLOB\_IGNORE] A list of shell patterns for filenames that \verb+glob+ should ignore.
 * \item[GLOB\_ALLOW] A list of shell pattern.  If a file does not match a pattern in
 *    \verb+GLOB_ALLOW+, it is ignored.
 * \end{description}
 *
 * The returned files are sorted by name.
 * \end{doc}
 *)
let glob venv pos loc args =
   let pos = string_pos "glob" pos in
   let option, arg =
      match args with
         [arg] ->
            ValString "", arg
       | [option; arg] ->
            option, arg
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))
   in
   let options = glob_options_of_env venv pos in
   let option = string_of_value venv pos option in
   let options = glob_options_of_string options option in
   let dirs = strings_of_value venv pos arg in
   let root = Dir.cwd () in
   let cwd = Dir.name root (venv_dir venv) in
   let dirs, names = glob options cwd dirs in
   let dirs = List.map (fun dir -> ValDir (Dir.chdir root dir)) dirs in
   let nodes = List.map (fun name -> ValNode (Node.intern root name)) names in
      ValArray (sort_val_nodes nodes)

(*
 * ls function.
 *
 * \begin{doc}
 * \subsection{ls}
 *
 * \begin{verbatim}
 *    $(ls files) : Node Array
 *       files : String Sequence
 *    $(ls options, files) : Node Array
 *       files : String Sequence
 * \end{verbatim}
 *
 * The \verb+ls+ function returns the filenames in a directory.
 *
 * The . and .. entries are always ignored.
 * The patterns are shell-style patterns, and are glob-expanded.
 *
 * The options include all of the options to the \verb+glob+ function,
 * plus the following.
 *
 * \begin{description}
 * \item[R] Perform a recursive listing.
 * \end{description}
 *
 * The \verb+GLOB_ALLOW+ and \verb+GLOB_IGNORE+ variables can be defined
 * to control the globbing behavior.
 * The returned files are sorted by name.
 * \end{doc}
 *)
let ls_fun_of_string s =
   let len = String.length s in
   let rec search i =
      if i = len then
         list_dirs
      else
         match s.[i] with
            'R' -> list_dirs_rec
          | _ -> search (succ i)
   in
      search 0

let ls venv pos loc args =
   let pos = string_pos "ls" pos in
   let option, arg =
      match args with
         [arg] ->
            ValString "", arg
       | [option; arg] ->
            option, arg
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))
   in
   let option = string_of_value venv pos option in
   let ls_fun  = ls_fun_of_string option in
   let options = glob_options_of_env venv pos in
   let options = glob_options_of_string options option in
   let dirs = strings_of_value venv pos arg in
   let root = Dir.cwd () in
   let cwd  = Dir.name root (venv_dir venv) in
   let dirs, files1 = Lm_glob.glob options cwd dirs in
   let dirs, files2 = ls_fun options dirs in
   let dirs  = List.map (fun dir -> ValDir (Dir.chdir root dir)) dirs in
   let nodes = List.map (fun name -> ValNode (Node.intern root name)) (files1 @ files2) in
   let nodes = dirs @ nodes in
      ValArray (sort_val_nodes nodes)

(*
 * The builtin function.
 *
 * \begin{doc}
 * \subsection{subdirs}
 *
 * \begin{verbatim}
 *    $(subdirs dirs) : Dir Array
 *       dirs : String Sequence
 *    $(subdirs options, dirs) : Dir Array
 *       options : String
 *       dirs : String Sequence
 * \end{verbatim}
 *
 * The \verb+subdirs+ function returns all the subdirectories
 * of a list of directories, recursively.
 *
 * The possible options are the following:
 * \begin{description}
 * \item[A] Return directories that begin with a .
 * \item[C] Ignore files according to \File{.cvsignore} rules.
 * \end{description}
 * \end{doc}
 *)
let subdirs venv pos loc args =
   let pos = string_pos "subdirs" pos in
   let option, arg =
      match args with
         [arg] ->
            ValString "", arg
       | [options; arg] ->
            options, arg
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))
   in
   let options = glob_options_of_env venv pos in
   let options = glob_options_of_string options (string_of_value venv pos option) in
   let dirs = strings_of_value venv pos arg in
   let root = Dir.cwd () in
   let cwd = Dir.name root (venv_dir venv) in
   let dirs = List.map (Filename.concat cwd) dirs in
   let dirs = subdirs_of_dirs options dirs in
   let dirs = List.map (fun dir -> ValDir (Dir.chdir root dir)) dirs in
      ValArray dirs

(************************************************************************
 * Directories.
 *)

(*
 * \begin{doc}
 * \subsection{mkdir}
 *
 * \begin{verbatim}
 *    mkdir(mode, node...)
 *       mode : Int
 *       node : Node
 *    raises RuntimeException
 * \end{verbatim}
 *
 * The \verb+mkdir+ function creates a directory.
 * \end{doc}
 *)
let mkdir venv pos loc args =
   let pos = string_pos "mkdir" pos in
      match args with
         [mode; nodes] ->
            let mode = int_of_value venv pos mode in
            let nodes = values_of_value venv pos nodes in
            let () =
               try
                  List.iter (fun node ->
                        Unix.mkdir (filename_of_value venv pos node) mode) nodes
               with
                  Unix.Unix_error _ as exn ->
                     raise (UncaughtException (pos, exn))
            in
               ValNone
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))

(************************************************************************
 * Stat.
 *)

(*
 * \begin{doc}
 * \subsection{Stat}
 *
 * The \verb+Stat+ object represents the result returned by the \verb+stat+
 * and \verb+lstat+ functions.  It contains the following fields.
 *
 * A \verb+stat+ object has the following fields.  Not all of the fields
 * will have meaning on all architectures.
 *
 * \begin{description}
 * \item[dev]: the device number.
 * \item[ino]: the inode number.
 * \item[kind]: the kind of the file, one of the following:
 *    \verb+REG+ (regular file),
 *    \verb+DIR+ (directory),
 *    \verb+CHR+ (character device),
 *    \verb+BLK+ (block device),
 *    \verb+LNK+ (symbolic link),
 *    \verb+FIFO+ (named pipe),
 *    \verb+SOCK+ (socket).
 * \item[perm]: access rights, represented as an integer.
 * \item[nlink]: number of links.
 * \item[uid]: user id of the owner.
 * \item[gid]: group id of the file's group.
 * \item[rdev]: device minor number.
 * \item[size]: size in bytes.
 * \item[atime]: last access time, as a floating point number.
 * \item[mtime]: last modification time, as a floating point number.
 * \item[ctime]: last status change time, as a floating point number.
 * \end{description}
 *
 * \subsection{stat}
 *
 * \begin{verbatim}
 *     $(stat node...) : Stat
 *        node : Node or Channel
 *     $(lstat node...) : Stat
 *        node : Node or Channel
 *     raises RuntimeException
 * \end{verbatim}
 *
 * The \verb+stat+ functions return file information.
 * If the file is a symbolic link, the \verb+stat+ function refers to the
 * destination of the link; the \verb+lstat+ function refers to the link
 * itself.
 * \end{doc}
 *)
let create_stat_obj obj stat =
   let obj = venv_add_field obj st_dev_sym (ValInt stat.Unix.st_dev) in
   let obj = venv_add_field obj st_ino_sym (ValInt stat.Unix.st_ino) in
   let kind =
      match stat.Unix.st_kind with
         Unix.S_REG -> "REG"
       | Unix.S_DIR -> "DIR"
       | Unix.S_CHR -> "CHR"
       | Unix.S_BLK -> "BLK"
       | Unix.S_LNK -> "LNK"
       | Unix.S_FIFO -> "FIFO"
       | Unix.S_SOCK -> "SOCK"
   in
   let obj = venv_add_field obj st_kind_sym (ValString kind) in
   let obj = venv_add_field obj st_perm_sym (ValInt stat.Unix.st_perm) in
   let obj = venv_add_field obj st_nlink_sym (ValInt stat.Unix.st_nlink) in
   let obj = venv_add_field obj st_uid_sym   (ValInt stat.Unix.st_uid) in
   let obj = venv_add_field obj st_gid_sym   (ValInt stat.Unix.st_gid) in
   let obj = venv_add_field obj st_rdev_sym  (ValInt stat.Unix.st_rdev) in
   let obj = venv_add_field obj st_size_sym  (ValInt stat.Unix.st_size) in
   let obj = venv_add_field obj st_atime_sym (ValFloat stat.Unix.st_atime) in
   let obj = venv_add_field obj st_mtime_sym (ValFloat stat.Unix.st_mtime) in
   let obj = venv_add_field obj st_ctime_sym (ValFloat stat.Unix.st_ctime) in
      ValObject obj

let stat_aux stat_fun venv pos loc args =
   let pos = string_pos "stat" pos in
   let obj =
      try
         match venv_find_var_exn venv ScopePublic stat_object_sym with
            ValObject obj ->
               obj
          | _ ->
               venv_empty_object
      with
         Not_found ->
            venv_empty_object
   in
      match args with
         [arg] ->
            let args = values_of_value venv pos arg in
            let stats =
               List.map (fun arg ->
                     let file = filename_of_value venv pos arg in
                     let stat =
                        try stat_fun file with
                           Unix.Unix_error _ as exn ->
                              raise (UncaughtException (pos, exn))
                     in
                        create_stat_obj obj stat) args
            in
               concat_array stats
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

let stat = stat_aux Unix.stat

let lstat =
   if Sys.os_type = "Win32" then
      stat_aux Unix.stat
   else
      stat_aux Unix.lstat

(************************************************************************
 * Links.
 *)

(*
 * \begin{doc}
 * \subsection{unlink}
 *
 * \begin{verbatim}
 *    $(unlink file...)
 *       file : File
 *    $(rmdir dir...)
 *       dir : Dir
 *    raises RuntimeException
 * \end{verbatim}
 *
 * The \verb+unlink+ function removes a file.
 * The \verb+rmdir+ function removes a directory.
 * \end{doc}
 *)
let unlink_aux rm_fun venv pos loc args =
   let pos = string_pos "unlink" pos in
      match args with
         [arg] ->
            let args = values_of_value venv pos arg in
            let () =
               try
                  List.iter (fun arg ->
                        rm_fun (filename_of_value venv pos arg)) args
               with
                  Unix.Unix_error _ as exn ->
                     raise (UncaughtException (pos, exn))
            in
               ValNone
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

let unlink = unlink_aux Unix.unlink
let rmdir = unlink_aux Unix.rmdir
let rm = unlink_aux (fun file ->
               try Unix.unlink file with
                  Unix.Unix_error _ ->
                     ())
(*
 * \begin{doc}
 * \subsection{rename}
 *
 * \begin{verbatim}
 *     rename(old, new)
 *        old : Node
 *        new : Node
 *     mv(nodes dir)
 *        nodes : Node Sequence
 *        dir   : Dir
 *     raises RuntimeException
 * \end{verbatim}
 *
 * The \verb+rename+ function changes the name of a file or directory named \verb+old+
 * to \verb+new+.
 *
 * The \verb+mv+ function is similar, but if \verb+new+ is a directory, and it exists,
 * then the files specified by the sequence are moved into the directory.  If not,
 * the behavior of \verb+mv+ is identical to \verb+rename+.
 * \end{doc}
 *)
let rename venv pos loc args =
   let pos = string_pos "rename" pos in
      match args with
         [node1; node2] ->
            let name1 = filename_of_value venv pos node1 in
            let name2 = filename_of_value venv pos node2 in
            let () =
               try Unix.rename name1 name2 with
                  Unix.Unix_error _ as exn ->
                     raise (UncaughtException (pos, exn))
            in
               ValNone
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))

let rec split_last l' l =
   match l with
      [h] ->
         List.rev l', h
    | h :: l ->
         split_last (h :: l') l
    | [] ->
         raise (Invalid_argument "split_last")

let mv_aux mv venv pos loc args =
   let pos = string_pos "mv" pos in
      match args with
         [args] ->
            let nodes = values_of_value venv pos args in
            let nodes, dir =
               match nodes with
                  [] ->
                     raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, 0)))
                | _ ->
                     split_last [] nodes
            in
            let dir = filename_of_value venv pos dir in
            let is_dir =
               try (Unix.stat dir).Unix.st_kind = Unix.S_DIR with
                  Unix.Unix_error _ as exn ->
                     false
            in
            let () =
               try
                  if is_dir then
                     List.iter (fun node ->
                           let name = filename_of_value venv pos node in
                              mv name (Filename.concat dir (Filename.basename name))) nodes
                  else
                     match nodes with
                        [node] ->
                           mv (filename_of_value venv pos node) dir
                      | _ ->
                           raise (OmakeException (loc_pos loc pos, StringStringError ("destination directory does not exist", dir)))
               with
                  Unix.Unix_error _ as exn ->
                     raise (UncaughtException (pos, exn))
            in
               ValNone
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

let mv = mv_aux Unix.rename

let cp_file file1 file2 =
   let fdr = Unix.openfile file1 [Unix.O_RDONLY] 0 in
   let fdw =
      try Unix.openfile file2 [Unix.O_WRONLY; Unix.O_CREAT; Unix.O_TRUNC] 0o666 with
         exn ->
            Unix.close fdr;
            raise exn
   in
      try
         let max = 4096 in
         let buffer = String.create max in
         let rec loop () =
            let count = Unix.read fdr buffer 0 max in
               if count <> 0 then
                  let rec write off =
                     if off < count then
                        write (off + Unix.write fdw buffer off (count - off))
                  in
                     write 0;
                     loop ()
         in
            loop ();
            Unix.close fdr;
            Unix.close fdw
      with
         exn ->
            Unix.close fdr;
            Unix.close fdw;
            raise exn

let cp = mv_aux cp_file

(*
 * \begin{doc}
 * \subsection{link}
 *
 * \begin{verbatim}
 *    link(src, dst)
 *       src : Node
 *       dst : Node
 *    raises RuntimeException
 * \end{verbatim}
 *
 * The \verb+link+ function creates a hard link named \verb+dst+ to the file
 * or directory \verb+src+.
 *
 * Hard links are not supported in Win32.
 *
 * Normally, only the superuser can create hard links to directories.
 * \end{doc}
 *)
let link venv pos loc args =
   let pos = string_pos "link" pos in
      match args with
         [src; dst] ->
            let src = filename_of_value venv pos src in
            let dst = filename_of_value venv pos dst in
            let () =
               try Unix.link src dst with
                  Unix.Unix_error _ as exn ->
                     raise (UncaughtException (pos, exn))
            in
               ValNone
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))

(*
 * \begin{doc}
 * \subsection{symlink}
 *
 * \begin{verbatim}
 *    symlink(src, dst)
 *       src : Node
 *       dst : Node
 *    raises RuntimeException
 * \end{verbatim}
 *
 * The \verb+symlink+ function creates a symbolic link \verb+dst+ that
 * points to the \verb+src+ file.
 *
 * The link name is computed relative to
 * the target directory.  For example, the expression
 * \verb+$(symlink a/b, c/d)+ creates a link named
 * \verb+c/d -> ../a/b+.
 *
 * Symbolic links are not suppoerted in Win32.
 * \end{doc}
 *)
let symlink venv pos loc args =
   let pos = string_pos "symlink" pos in
      match args with
         [src; dst] ->
            let dst = file_of_value venv pos dst in
            let src = file_of_value venv pos src in
            let src = Node.name (Node.dir dst) src in
            let dst = Node.fullname dst in
            let () =
               try Unix.symlink src dst with
                  Unix.Unix_error _ as exn ->
                     raise (UncaughtException (pos, exn))
            in
               ValNone
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))

(*
 * \begin{doc}
 * \subsection{readlink}
 *
 * \begin{verbatim}
 *    $(readlink node...) : Node
 *       node : Node
 * \end{verbatim}
 *
 * The \verb+readlink+ function reads the value of a symbolic link.
 * \end{doc}
 *)
let readlink venv pos loc args =
   let pos = string_pos "readlink" pos in
      match args with
         [arg] ->
            let args = values_of_value venv pos arg in
            let args =
               try
                  List.map (fun arg ->
                        let node = file_of_value venv pos arg in
                        let dir = Node.dir node in
                        let name = Node.fullname node in
                        let name = Unix.readlink name in
                           ValNode (Node.intern dir name)) args
               with
                  Unix.Unix_error _ as exn ->
                     raise (UncaughtException (pos, exn))
            in
               concat_array args
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(************************************************************************
 * Permissions.
 *)

(*
 * \begin{doc}
 * \subsection{chmod}
 *
 * \begin{verbatim}
 *    chmod(mode, dst...)
 *       mode : Int
 *       dst : Node or Channel
 *    raises RuntimeException
 * \end{verbatim}
 *
 * The \verb+chmod+ function changes the permissions of the targets.
 * \end{doc}
 *)
let chmod venv pos loc args =
   let pos = string_pos "chmod" pos in
      match args with
         [mode; arg] ->
            let mode = int_of_value venv pos mode in
            let args = values_of_value venv pos arg in
            let () =
               try
                  List.iter (fun arg ->
                        Unix.chmod (filename_of_value venv pos arg) mode) args
               with
                  Unix.Unix_error _ as exn ->
                     raise (UncaughtException (pos, exn))
            in
               ValNone
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 2, List.length args)))

(*
 * \begin{doc}
 * \subsection{chown}
 *
 * \begin{verbatim}
 *    chown(uid, gid, node...)
 *       uid : Int
 *       gid : Int
 *       node : Node or Channel
 *    chown(uid, node...)
 *       uid : Int
 *       node : Node or Channel
 *    raises RuntimeException
 * \end{verbatim}
 *
 * The \verb+chown+ function changes the user and group id of the file.
 * If the \verb+gid+ is not specified, it is not changed.  If either
 * id is -1, that id is not changed.
 * \end{doc}
 *)
let chown venv pos loc args =
   let pos = string_pos "chown" pos in
   let uid, gid, nodes =
      match args with
         [uid; nodes] ->
            uid, None, nodes
       | [uid; gid; nodes] ->
            uid, Some gid, nodes
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityRange (2, 3), List.length args)))
   in
   let uid = int_of_value venv pos uid in
   let gid =
      match gid with
         Some gid ->
            int_of_value venv pos gid
       | None ->
            -1
   in
   let nodes = values_of_value venv pos nodes in
   let () =
      try
         List.iter (fun node ->
               Unix.chown (filename_of_value venv pos node) uid gid) nodes
      with
         Unix.Unix_error _ as exn ->
            raise (UncaughtException (pos, exn))
   in
      ValNone

(*
 * \begin{doc}
 * \subsection{umask}
 *
 * \begin{verbatim}
 *     $(umask mode) : Int
 *        mode : Int
 *     raises RuntimeException
 * \end{verbatim}
 *
 * Sets the file mode creation mask.
 * The previous mask is returned.
 * This value is not scoped, changes have global effect.
 * \end{doc}
 *)
let umask venv pos loc args =
   let pos = string_pos "umask" pos in
      match args with
         [arg] ->
            let mode = int_of_value venv pos arg in
            let mask =
               try Unix.umask mode with
                  Unix.Unix_error _ as exn ->
                     raise (UncaughtException (pos, exn))
            in
               ValInt mask
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(************************************************************************
 * Tables.
 *)

(*
 * Table of variables.
 *)
let builtin_vars =
   []

(*
 * Table of built-in functions.
 *)
let builtin_funs =
   [true, "file-sort",             sort,                ArityExact 2;
    true, "file-check-sort",       check_sort,          ArityExact 2;
    true, "ls",                    ls,                  ArityRange (1, 2);
    true, "glob",                  glob,                ArityRange (1, 2);
    true, "subdirs",               subdirs,             ArityRange (1, 2);
    true, "basename",              basename,            ArityExact 1;
    true, "homename",              homename,            ArityExact 1;
    true, "rootname",              rootname,            ArityExact 1;
    true, "suffix",                suffix,              ArityExact 1;
    true, "file",                  file,                ArityExact 1;
    true, "dir",                   dir,                 ArityExact 1;
    true, "which",                 which,               ArityExact 1;
    true, "exists-in-path",        exists_in_path,      ArityExact 1;
    true, "in",                    ind,                 ArityExact 2;
    true, "dirof",                 dirof,               ArityExact 1;
    true, "file-exists",           file_exists,         ArityExact 1;
    true, "filter-exists",         filter_exists,       ArityExact 1;
    true, "target-exists",         target_exists,       ArityExact 1;
    true, "filter-targets",        filter_targets,      ArityExact 1;
    true, "mkdir",                 mkdir,               ArityExact 2;
    true, "stat",                  stat,                ArityExact 1;
    true, "lstat",                 lstat,               ArityExact 1;
    true, "unlink",                unlink,              ArityExact 1;
    true, "rmdir",                 rmdir,               ArityExact 1;
    true, "rename",                rename,              ArityExact 2;
    true, "mv",                    mv,                  ArityExact 2;
    true, "link",                  link,                ArityExact 2;
    true, "symlink",               symlink,             ArityExact 2;
    true, "readlink",              readlink,            ArityExact 1;
    true, "chmod",                 chmod,               ArityExact 2;
    true, "chown",                 chown,               ArityRange (2, 3);
    true, "umask",                 umask,               ArityExact 1;
    true, "cp",                    cp,                  ArityExact 1;
    true, "rm",                    rm,                  ArityExact 1]

(*
 * Builtin rules.
 *)
let phony_targets = []

let builtin_rules =
   []

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
