(*
 * Convert the AST to an IR representation.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_printf
open Lm_location
open Lm_symbol

open Omake_ir
open Omake_env
open Omake_symbol
open Omake_node_sig

module Pos = MakePos (struct let name = "Omake_ir_ast" end)
open Pos

(*
 * Temporary type for collecting try blocks.
 *)
type ast_exp = Omake_ast.exp

type senv =
   { senv_class_names : symbol list;
     senv_mode        : scope_kind;
     senv_vars        : scope_kind SymbolTable.t
   }

(************************************************************************
 * Utilities.
 *)

(*
 * All current Protected vars become Static.
 *)
let object_vars vars =
   SymbolTable.map (function
      ScopeProtected ->
         ScopePrivate
    | kind ->
         kind) vars

(*
 * Collect the cases in a conditional.
 *)
let rec collect_if cases el =
   match el with
      Omake_ast.CommandExp (v, e, body, loc) :: el when Lm_symbol.eq v elseif_sym ->
         collect_if ((e, body) :: cases) el
    | Omake_ast.CommandExp (v, e, body, loc) :: el when Lm_symbol.eq v else_sym ->
         let cases = (Omake_ast.StringExp ("true", loc), body) :: cases in
            List.rev cases, el
    | _ ->
         List.rev cases, el

(*
 * Generic case collection.
 *)
let rec collect_cases cases el =
   match el with
      (Omake_ast.CommandExp (v, e, body, _) :: el) when SymbolSet.mem clauses_set v ->
         collect_cases ((v, e, body) :: cases) el
    | (Omake_ast.CatchExp (v1, v2, body, loc) :: el) ->
         collect_cases ((v1, Omake_ast.StringExp (Lm_symbol.to_string v2, loc), body) :: cases) el
    | _ ->
         List.rev cases, el

(************************************************************************
 * Scoping.
 *)

(*
 * Default is public.
 *)
let senv_empty =
   { senv_class_names = [];
     senv_mode        = ScopePublic;
     senv_vars        = SymbolTable.empty
   }

let senv_class_names senv =
   senv.senv_class_names

(*
 * Change the scope mode.
 *)
let senv_scope senv scope =
   let mode = senv.senv_mode in
   let senv = { senv with senv_mode = scope } in
      senv, mode

(*
 * Get the scope for a variable.
 *)
let senv_find_var senv v =
   try SymbolTable.find senv.senv_vars v with
      Not_found ->
         ScopePublic

let senv_find_method_var senv pos vl =
   match vl with
      v :: _ ->
         senv_find_var senv v
    | [] ->
         raise (OmakeException (pos, StringError "empty method name"))

let senv_add_var senv v =
   let { senv_mode = scope;
         senv_vars = vars
       }  = senv
   in
      try senv, SymbolTable.find senv.senv_vars v with
         Not_found ->
            { senv with senv_vars = SymbolTable.add vars v scope }, scope

let senv_add_params senv params =
   let { senv_mode = scope;
         senv_vars = vars
       } = senv
   in
      { senv with senv_vars = List.fold_left (fun vars v -> SymbolTable.add vars v scope) vars params }, scope

let senv_add_method_var senv pos loc vl =
   match vl with
      [v] ->
         let senv, scope = senv_add_var senv v in
            senv, scope, v
    | [scope_var; v] ->
         let scope =
            if Lm_symbol.eq scope_var this_sym || Lm_symbol.eq scope_var protected_sym then
               ScopeProtected
            else if Lm_symbol.eq scope_var private_sym then
               ScopePrivate
            else if Lm_symbol.eq scope_var public_sym then
               ScopePublic
            else
               raise (OmakeException (loc_pos loc pos, StringMethodError ("illegal definition", vl)))
         in
            senv, scope, v
    | _ ->
         raise (OmakeException (loc_pos loc pos, StringMethodError ("illegal definition", vl)))

(************************************************************************
 * Conversion
 *)

(*
 * Conversion.
 *)
let rec build_string senv e pos =
   let pos = string_pos "build_string" pos in
      match e with
         Omake_ast.NullExp loc ->
            ConstString (loc, "")
       | Omake_ast.StringExp (s, loc) ->
            ConstString (loc, s)
       | Omake_ast.QuoteExp (el, loc) ->
            build_quote_string senv el pos loc
       | Omake_ast.SequenceExp ([e], _) ->
            build_string senv e pos
       | Omake_ast.SequenceExp (el, loc) ->
            build_sequence_string senv el pos loc
       | Omake_ast.ApplyExp (strategy, v, args, loc) ->
            build_apply_string senv strategy v args pos loc
       | Omake_ast.SuperApplyExp (strategy, super, v, args, loc) ->
            build_super_apply_string senv strategy super v args pos loc
       | Omake_ast.MethodApplyExp (strategy, vl, args, loc) ->
            build_method_apply_string senv strategy vl args pos loc
       | Omake_ast.BodyExp (el, loc) ->
            build_body_string senv el pos loc
       | Omake_ast.CommandExp (_, _, _, loc)
       | Omake_ast.VarDefExp (_, _, _, _, loc)
       | Omake_ast.VarDefBodyExp (_, _, _, _, loc)
       | Omake_ast.ObjectDefExp (_, _, _, loc)
       | Omake_ast.FunDefExp (_, _, _, loc)
       | Omake_ast.RuleExp (_, _, _, _, _, loc)
       | Omake_ast.ShellExp (_, loc)
       | Omake_ast.CatchExp (_, _, _, loc)
       | Omake_ast.ClassExp (_, loc) ->
            raise (OmakeException (loc_pos loc pos, SyntaxError "misplaced expression"))

and build_string_list senv el pos =
   let pos = string_pos "build_string_list" pos in
      List.map (fun e -> build_string senv e pos) el

and build_string_opt senv sl pos =
   let pos = string_pos "build_string_opt" pos in
      match sl with
         Some s ->
            Some (build_string senv s pos)
       | None ->
            None

(*
 * When building a sequence, try to collapse adjacent constant strings.
 *)
and build_sequence_string senv el pos loc =
   let pos = string_pos "build_sequence_string" pos in
   let args = build_sequence_string_aux senv el pos loc in
      SequenceString (loc, args)

and build_quote_string senv el pos loc =
   let pos = string_pos "build_quote_string" pos in
   let args = build_sequence_string_aux senv el pos loc in
      QuoteString (loc, args)

and build_sequence_string_aux senv el pos loc =
   let pos = string_pos "build_sequence_string_aux" pos in
   let buf = Buffer.create 32 in

   (* Flush the buffer *)
   let flush_buffer buf_opt args =
      match buf_opt with
         Some loc ->
            let args = ConstString (loc, Buffer.contents buf) :: args in
               Buffer.clear buf;
               args
       | None ->
            args
   in

   (* Add a constant string to the buffer *)
   let add_string buf_opt s loc =
      Buffer.add_string buf s;
      match buf_opt with
         Some loc' ->
            let loc = union_loc loc' loc in
               Some loc
       | None ->
            Some loc
   in

   (* Collect all the strings in the sequence *)
   let rec collect buf_opt args el =
      match el with
         [] ->
            let args = flush_buffer buf_opt args in
               List.rev args
       | e :: el ->
            let e = build_string senv e pos in
               match e with
                  ConstString (loc, s) ->
                     let buf_opt = add_string buf_opt s loc in
                        collect buf_opt args el
                | ApplyString _
                | SuperApplyString _
                | MethodApplyString _
                | SequenceString _
                | BodyString _
                | ArrayString _
                | ArrayOfString _
                | QuoteString _
                | ExpString _
                | CasesString _
                | ThisString _ ->
                     let args = flush_buffer buf_opt args in
                     let args = e :: args in
                        collect None args el
   in
      collect None [] el

(*
 * Build an application.
 *)
and build_apply_string senv strategy v args pos loc =
   let pos = string_pos "build_apply_string" pos in
   let strategy =
      match strategy with
         Omake_ast.LazyApply ->
            LazyApply
       | Omake_ast.EagerApply ->
            EagerApply
       | Omake_ast.NormalApply ->
            NormalApply
   in
      if Lm_symbol.eq v this_sym || Lm_symbol.eq v protected_sym then
         ThisString (loc, ScopeProtected)
      else if Lm_symbol.eq v public_sym then
         ThisString (loc, ScopePublic)
      else if Lm_symbol.eq v private_sym then
         ThisString (loc, ScopePrivate)
      else
         let scope = senv_find_var senv v in
         let args = build_string_list senv args pos in
            ApplyString (loc, strategy, scope, v, args)

(*
 * Super call.
 *)
and build_super_apply_string senv strategy super v args pos loc =
   let pos = string_pos "build_super_apply_string" pos in
   let strategy =
      match strategy with
         Omake_ast.LazyApply ->
            LazyApply
       | Omake_ast.EagerApply ->
            EagerApply
       | Omake_ast.NormalApply ->
            NormalApply
   in
   let scope = senv_find_var senv super in
   let args = build_string_list senv args pos in
      SuperApplyString (loc, strategy, scope, super, v, args)

(*
 * Build a method application.
 *)
and build_method_apply_string senv strategy vl args pos loc =
   let pos = string_pos "build_method_apply_string" pos in
   let scope, vl =
      match vl with
         v :: vl' ->
            if Lm_symbol.eq v this_sym || Lm_symbol.eq v protected_sym then
               ScopeProtected, vl'
            else if Lm_symbol.eq v public_sym then
               ScopePublic, vl'
            else if Lm_symbol.eq v private_sym then
               ScopePrivate, vl'
            else
               senv_find_var senv v, vl
       | [] ->
            raise (OmakeException (loc_pos loc pos, StringError "empty method name"))
   in
   let args = build_string_list senv args pos in
   let strategy =
      match strategy with
         Omake_ast.LazyApply ->
            LazyApply
       | Omake_ast.EagerApply ->
            EagerApply
       | Omake_ast.NormalApply ->
            NormalApply
   in
      match vl with
         [v] ->
            ApplyString (loc, strategy, scope, v, args)
       | _ ->
            MethodApplyString (loc, strategy, scope, vl, args)

(*
 * Build a body expression.
 *)
and build_body_string senv el pos loc =
   let pos = string_pos "build_body_string" pos in
   let _, body = build_sequence_exp senv el pos loc in
      BodyString (loc, body)

(*
 * Build an expression.
 *)
and build_exp senv e =
   let pos = string_pos "build_exp" (ast_exp_pos e) in
      match e with
         Omake_ast.NullExp loc
       | Omake_ast.StringExp (_, loc)
       | Omake_ast.QuoteExp (_, loc) ->
            senv, SequenceExp (loc, [])
       | Omake_ast.SequenceExp ([e], _)
       | Omake_ast.BodyExp ([e], _) ->
            build_exp senv e
       | Omake_ast.SequenceExp (el, loc)
       | Omake_ast.BodyExp (el, loc) ->
            build_sequence_exp senv el pos loc
       | Omake_ast.ApplyExp (_, v, args, loc) ->
            build_apply_exp senv v args pos loc
       | Omake_ast.SuperApplyExp (_, super, v, args, loc) ->
            build_super_apply_exp senv super v args pos loc
       | Omake_ast.MethodApplyExp (_, vl, args, loc) ->
            build_method_apply_exp senv vl args pos loc
       | Omake_ast.CommandExp (v, arg, commands, loc) ->
            build_command_exp senv v arg commands pos loc
       | Omake_ast.VarDefExp (v, kind, flag, e, loc) ->
            build_var_def_exp senv v kind flag e pos loc
       | Omake_ast.VarDefBodyExp (v, kind, flag, [], loc) ->
            build_var_def_exp senv v kind flag (Omake_ast.SequenceExp ([], loc)) pos loc
       | Omake_ast.VarDefBodyExp (v, kind, flag, el, loc) ->
            build_var_def_body_exp senv v kind flag el pos loc
       | Omake_ast.ObjectDefExp (v, flag, el, loc) ->
            build_object_def_exp senv v flag el pos loc
       | Omake_ast.FunDefExp (v, params, e, loc) ->
            build_fun_def_exp senv v params e pos loc
       | Omake_ast.RuleExp (multiple, target, pattern, source, commands, loc) ->
            build_rule_exp senv multiple target pattern source commands pos loc
       | Omake_ast.ShellExp (e, loc) ->
            build_shell_exp senv e pos loc
       | Omake_ast.CatchExp (_, _, _, loc) ->
            raise (OmakeException (pos, StringError "misplaced catch clause"))
       | Omake_ast.ClassExp (names, loc) ->
            build_class_exp senv loc names

(*
 * Add the class names.
 *)
and build_class_exp senv loc names =
   let senv =
      { senv with senv_mode = ScopeProtected;
                  senv_class_names = names @ senv.senv_class_names
      }
   in
      senv, SequenceExp (loc, [])

(*
 * Sequence exp.  Build the expression one at a time.
 *)
and build_sequence senv pos loc rval el =
   match el with
      Omake_ast.CommandExp (v, e, body, _)
      :: el when Lm_symbol.eq v if_sym ->
         let cases, el = collect_if [e, body] el in
         let senv, e = build_if_exp senv cases pos loc in
         let senv, el = build_sequence senv pos loc rval el in
            senv, e :: el

    | Omake_ast.CommandExp (v, arg, body, _) :: el ->
         let cases, el = collect_cases [] el in
         let senv, e = build_cases_command_exp senv v arg cases body pos loc in
         let senv, el = build_sequence senv pos loc rval el in
            senv, e :: el

    | Omake_ast.ApplyExp (_, v, args, loc) :: el ->
         let cases, el = collect_cases [] el in
         let senv, e = build_cases_apply_exp senv v args cases pos loc in
         let senv, el = build_sequence senv pos loc rval el in
            senv, e :: el

    | e :: el ->
         let senv, e = build_exp senv e in
         let senv, el = build_sequence senv pos loc rval el in
            senv, e :: el

    | [] ->
         rval senv

and build_sequence_exp senv el pos loc =
   let pos = string_pos "build_sequence_exp" pos in
   let _, el = build_sequence senv pos loc (fun senv -> senv, []) el in
      senv, SequenceExp (loc, el)

and build_scope_exp senv scope el pos loc =
   let pos = string_pos "build_scope_exp" pos in
   let senv, scope = senv_scope senv scope in
   let senv, el = build_sequence senv pos loc (fun senv -> senv, []) el in
   let senv, _ = senv_scope senv scope in
      senv, SequenceExp (loc, el)

(*
 * In an application, turn the arguments into strings.
 *)
and build_apply_exp senv v args pos loc =
   let pos = string_pos "build_apply_exp" pos in
   let args = build_string_list senv args pos in
   let scope = senv_find_var senv v in
      senv, ApplyExp (loc, scope, v, args)

and build_super_apply_exp senv super v args pos loc =
   let pos = string_pos "build_super_apply_exp" pos in
   let args = build_string_list senv args pos in
   let scope = senv_find_var senv super in
      senv, SuperApplyExp (loc, scope, super, v, args)

and build_method_apply_exp senv vl args pos loc =
   let pos = string_pos "build_method_apply_exp" pos in
   let args = build_string_list senv args pos in
   let scope, vl =
      match vl with
         v :: vl' ->
            if Lm_symbol.eq v this_sym || Lm_symbol.eq v protected_sym then
               ScopeProtected, vl'
            else if Lm_symbol.eq v public_sym then
               ScopePublic, vl'
            else if Lm_symbol.eq v private_sym then
               ScopePrivate, vl'
            else
               senv_find_var senv v, vl
       | [] ->
            raise (OmakeException (loc_pos loc pos, StringError "empty method name"))
   in
   let e =
      match vl with
         [v] ->
            ApplyExp (loc, scope, v, args)
       | _ ->
            MethodApplyExp (loc, scope, vl, args)
   in
      senv, e

and build_cases_apply_exp senv v args cases pos loc =
   let pos = string_pos "build_cases_apply_exp" pos in
   let args = build_string_list senv args pos in
   let scope = senv_find_var senv v in
   let args =
      match cases with
         [] ->
            args
       | _ ->
            let cases =
               List.map (fun (v, e, body) ->
                     let _, body = build_sequence_exp senv body pos loc in
                        v, build_string senv e pos, body) cases
            in
               CasesString (loc, cases) :: args
   in
      senv, ApplyExp (loc, scope, v, args)

and build_cases_command_exp senv v arg cases commands pos loc =
   let pos = string_pos "build_cases_command_exp" pos in
      match cases with
         [] ->
            build_command_exp senv v arg commands pos loc
       | _ ->
            let arg =
               match commands with
                  [] ->
                     arg
                | _ ->
                     Omake_ast.BodyExp (commands, loc)
            in
               build_cases_apply_exp senv v [arg] cases pos loc

(*
 * Commands.
 *)
and build_command_exp senv v arg commands pos loc =
   let pos = string_pos "build_command_exp" pos in
      if Lm_symbol.eq v include_sym then
         build_include_exp senv arg commands pos loc
      else if Lm_symbol.eq v if_sym then
         build_if_exp senv [arg, commands] pos loc
      else if Lm_symbol.eq v section_sym then
         build_section_exp senv arg commands pos loc
      else if Lm_symbol.eq v private_sym then
         build_scope_exp senv ScopePrivate commands pos loc
      else if Lm_symbol.eq v protected_sym then
         build_scope_exp senv ScopeProtected commands pos loc
      else if Lm_symbol.eq v public_sym then
         build_scope_exp senv ScopePublic commands pos loc
      else if commands <> [] then
         raise (OmakeException (loc_pos loc pos, StringVarError ("illegal body for", v)))
      else
         build_apply_exp senv v [arg] pos loc

(*
 * Include a file.
 *)
and build_include_exp senv e commands pos loc =
   let pos = string_pos "build_include_exp" pos in
   let s = build_string senv e pos in
   let commands = build_string_list senv commands pos in
      senv, IncludeExp (loc, s, commands)

(*
 * Conditionals.
 *)
and build_if_exp senv cases pos loc =
   let pos = string_pos "build_if_exp" pos in
   let cases =
      List.map (fun (e1, e2) ->
            let s = build_string senv e1 pos in
            let _, e2 = build_sequence_exp senv e2 pos loc in
               s, e2) cases
   in
      senv, IfExp (loc, cases)

(*
 * A section is just an "if true" command.
 *)
and build_section_exp senv e1 e2 pos loc =
   let pos = string_pos "build_section_exp" pos in
   let s = build_string senv e1 pos in
   let _, body = build_sequence senv pos loc (fun senv -> senv, []) e2 in
      senv, SectionExp (loc, s, body)

(*
 * An object is a collection of definitions.
 *)
and build_object_def_exp senv v flag body pos loc =
   let pos = string_pos "build_object_def_exp" pos in
   let senv, scope, v = senv_add_method_var senv pos loc v in
   let pscope, parent =
      match flag with
         Omake_ast.DefineNormal ->
            ScopePublic, object_sym
       | Omake_ast.DefineAppend ->
            scope, v
   in
   let rval senv =
      senv, [ReturnObjectExp (loc, senv.senv_class_names)]
   in
   let senv' =
      { senv with senv_mode = ScopeProtected;
                  senv_vars = object_vars senv.senv_vars;
                  senv_class_names = []
      }
   in
   let _, body = build_sequence senv' pos loc rval body in
   let escope = senv_find_var senv extends_sym in
   let body = ApplyExp (loc, escope, extends_sym, [ApplyString (loc, EagerApply, pscope, parent, [])]) :: body in
      senv, LetObjectExp (loc, scope, v, body)

(*
 * Variable definition.
 *)
and build_var_def_kind flag =
   match flag with
      Omake_ast.DefineNormal ->
         VarDefNormal
    | Omake_ast.DefineAppend ->
         VarDefAppend

and build_var_def_exp senv v kind flag e pos loc =
   let pos = string_pos "build_var_def_exp" pos in
   let s = build_string senv e pos in
   let s =
      match kind with
         Omake_ast.DefineString ->
            s
       | Omake_ast.DefineArray ->
            ArrayOfString (loc, s)
   in
   let kind = build_var_def_kind flag in
      match v with
         [v] when Lm_symbol.eq v this_sym ->
            senv, LetThisExp (loc, s)
       | _ ->
            let senv, scope, v = senv_add_method_var senv pos loc v in
               senv, LetVarExp (loc, scope, v, kind, s)

(*
 * Variable definition.
 *)
and build_var_def_body_exp senv v kind flag body pos loc =
   let pos = string_pos "build_var_def_body_exp" pos in
   let e =
      match kind with
         Omake_ast.DefineString ->
            let _, el = build_sequence_exp senv body pos loc in
               ExpString (loc, el)
       | Omake_ast.DefineArray ->
            ArrayString (loc, build_string_list senv body pos)
   in
   let kind = build_var_def_kind flag in
   let senv, scope, v = senv_add_method_var senv pos loc v in
      senv, LetVarExp (loc, scope, v, kind, e)

(*
 * Function definition.
 *)
and build_fun_def_exp senv v params el pos loc =
   let pos = string_pos "build_fun_def_exp" pos in
   let senv', _ = senv_scope senv ScopePublic in
   let senv', _ = senv_add_params senv' params in
   let _, body = build_sequence_exp senv' el pos loc in
   let senv, scope, v = senv_add_method_var senv pos loc v in
      senv, LetFunExp (loc, scope, v, params, body)

(*
 * Special rule expressions.
 *)
and build_rule_exp senv multiple target pattern sources body pos loc =
   let pos = string_pos "build_rule_exp" pos in
   let target  = build_string senv target pos in
   let pattern = build_string senv pattern pos in

   (* Get the body *)
   let body =
      let senv = { senv with senv_class_names = [] } in
      let _, body = build_sequence senv pos loc (fun senv -> senv, []) body in
         BodyString (loc, SequenceExp (loc, body))
   in

   (* Get the option list *)
   let options =
      SymbolTable.fold (fun options v source ->
            if Lm_symbol.eq v normal_sym then
               options
            else
               let option =
                  ArrayString (loc, [ConstString (loc, Lm_symbol.to_string v); build_string senv source pos])
               in
                  option :: options) [] sources
   in
   let options = ArrayString (loc, options) in

   (* Add the sources *)
   let source =
      try
         let source = SymbolTable.find sources normal_sym in
            build_string senv source pos
      with
         Not_found ->
            ConstString (loc, "")
   in

   (* Multiple string *)
   let multiple =
      let s =
         if multiple then
            "true"
         else
            "false"
      in
         ConstString (loc, s)
   in

   (* Construct the argument list *)
   let args = [multiple; target; pattern; source; options; body] in
   let exp = ApplyExp (loc, ScopeProtected, rule_sym, args) in
      senv, exp

(*
 * Shell command.
 *)
and build_shell_exp senv e pos loc =
   let pos = string_pos "build_shell_exp" pos in
      senv, ShellExp (loc, build_string senv e pos)

(*
 * Build the IR from the AST program.
 *)
let compile_exp = build_exp

let compile_prog senv el =
   let loc = bogus_loc "Omake_ir_ast" in
   let pos = string_pos "compile" (loc_exp_pos loc) in
   let senv, el = build_sequence senv pos loc (fun senv -> senv, [ReturnSaveExp loc]) el in
      senv, SequenceExp (loc, el)

let compile_exp_list senv el =
   let loc = bogus_loc "Omake_ir_ast" in
   let pos = string_pos "compile" (loc_exp_pos loc) in
   let senv, el = build_sequence senv pos loc (fun senv -> senv, []) el in
      senv, SequenceExp (loc, el)

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
