/*
 * Read input from the terminal.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2004 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 */
#include <stdio.h>
#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>
#include <caml/fail.h>
#include <caml/custom.h>

/*
 * Maximum number of characters in an input line.
 */
#define MAX_LINE_LENGTH         256

/*
 * Prompts shouldn't be too wild.
 * In particular, they should be less than an average line length.
 */
#define MAX_PROMPT_LENGTH       80

#ifdef WIN32
#include <windows.h>

/*
 * Number of fields in the history.
 */
#define MAX_HISTORY_LENGTH      256

/*
 * Console modes.
 */
#define RAW_CONSOLE_MODE        (ENABLE_WINDOW_INPUT)
#define COOKED_CONSOLE_MODE     (ENABLE_ECHO_INPUT \
                                | ENABLE_LINE_INPUT\
                                | ENABLE_MOUSE_INPUT\
                                | ENABLE_PROCESSED_INPUT\
                                | ENABLE_WINDOW_INPUT)

/*
 * Number of events to read from the console.
 */
#define INPUT_COUNT             10


/*
 * A ReadLine-style buffer.
 * The buffer should be long enough to hold the input,
 * and the line terminator, and the string terminator.
 */
typedef struct _line_buffer {
    char buffer[MAX_LINE_LENGTH + 3];
    char length, index;
} LineBuffer;

typedef struct _readline {
    /* History */
    LineBuffer history[MAX_HISTORY_LENGTH];
    LineBuffer current;                 // The line being edited
    LineBuffer display;                 // What is displayed
    int start;                          // First line in the history
    int length;                         // Number of lines in the history
    int index;                          // Current line in the history

    /* Input processing */
    int escape;                         // Should next char be escaped?

    /* Prompts */
    char prompt[MAX_PROMPT_LENGTH];
    int prompt_length;

    /* Console information */
    int is_console;                     // Are we connected to a console?
    int force_prompt;                   // Force the prompt even if not a console
    HANDLE console_in;                  // Input console
    HANDLE console_out;                 // Screen buffer
    int x, y;                           // Original cursor position
    int w, h;                           // Total size of the window
} ReadLine;

/*
 * The codes for each operation.
 */
typedef enum {
    CODE_SUCCESS,
    CODE_EOL,
    CODE_EOF
} ProcessCode;

/************************************************************************
 * Utilities.
 */
/*
 * Print an error message.
 */
static void print_error_code(const char *name, DWORD code)
{
    LPTSTR buffer;

    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,  // Format flags
                  NULL,                                                         // Location of the message
                  code,                                                         // Error code
                  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),                    // Language
                  (LPTSTR) &buffer,                                             // Message buffer
                  0,                                                            // Buffer size
                  NULL);                                                        // Arguments

    /* Print the message */
    fprintf(stderr, "%s: failed with code %d: %s\n", name, errno, buffer);
    fflush(stderr);
    LocalFree(buffer);
}

static void print_error(const char *name)
{
    print_error_code(name, GetLastError());
}

/************************************************************************
 * Operations on LineBuffers
 */

/*
 * Index positions.
 */
static int IndexOfNextWord(LineBuffer *linep)
{
    int index;

    index = linep->index;
    while(index < linep->length && linep->buffer[index] > ' ')
        index++;
    while(index < linep->length && linep->buffer[index] <= ' ')
        index++;
    return index;
}

static int IndexOfPrevWord(LineBuffer *linep)
{
    int index;

    index = linep->index;
    while(index && linep->buffer[index] <= ' ')
        index--;
    while(index && linep->buffer[index] > ' ')
        index--;
    return index;
}

/*
 * Move the cursor.
 */
static void LineMoveNextChar(LineBuffer *linep)
{
    if(linep->index != linep->length)
        linep->index++;
}

static void LineMovePrevChar(LineBuffer *linep)
{
    if(linep->index)
        linep->index--;
}

static void LineMoveNextWord(LineBuffer *linep)
{
    linep->index = IndexOfNextWord(linep);
}

static void LineMovePrevWord(LineBuffer *linep)
{
    linep->index = IndexOfPrevWord(linep);
}

/*
 * Insert a char at the current position.
 */
static void LineInsertChar(LineBuffer *linep, char c)
{
    int amount;

    if(linep->index == MAX_LINE_LENGTH)
        return;

    /* Insert it */
    amount = linep->length - linep->index;
    memmove(linep->buffer + linep->index + 1, linep->buffer + linep->index, amount);
    linep->buffer[linep->index] = c;
    linep->index++;
    linep->length++;
}

/*
 * Delete characters between the current and the new index.
 */
static void LineDeleteRange(LineBuffer *linep, int index)
{
    int amount, remaining, tmp;

    if(index < linep->index) {
        tmp = index;
        index = linep->index;
        linep->index = tmp;
    }
    amount = index - linep->index;
    remaining = linep->length - index;
    memmove(linep->buffer + linep->index, linep->buffer + index, amount);
    linep->length -= amount;
}

static void LineDeletePrevChar(LineBuffer *linep)
{
    if(linep->index)
        LineDeleteRange(linep, linep->index - 1);
}

static void LineDeleteNextChar(LineBuffer *linep)
{
    if(linep->index < linep->length)
        LineDeleteRange(linep, linep->index + 1);
    else if(linep->index)
        LineDeleteRange(linep, linep->index - 1);
}

static void LineDeletePrevWord(LineBuffer *linep)
{
    LineDeleteRange(linep, IndexOfPrevWord(linep));
}

static void LineDeleteNextWord(LineBuffer *linep)
{
    LineDeleteRange(linep, IndexOfNextWord(linep));
}

static void LineDeleteToBOL(LineBuffer *linep)
{
    LineDeleteRange(linep, 0);
}

static void LineDeleteToEOL(LineBuffer *linep)
{
    linep->length = linep->index;
}

static void LineDeleteLine(LineBuffer *linep)
{
    linep->index = 0;
    linep->length = 0;
}

/*
 * Insert an EOL.
 */
static void LineInsertEOL(LineBuffer *linep)
{
    linep->buffer[linep->length] = '\r';
    linep->buffer[linep->length + 1] = '\n';
    linep->length += 2;
    linep->index = linep->length;
}

/************************************************************************
 * ReadLine operations.
 */

/*
 * Allocate a ReadLine struct.
 */
static ReadLine *AllocReadLine(int is_console, HANDLE console_in, HANDLE console_out)
{
    ReadLine *readp;

    /* Allocate and zero */
    readp = (ReadLine *) malloc(sizeof(ReadLine));
    if(readp) {
        memset(readp, 0, sizeof(ReadLine));
        readp->is_console = is_console;
        readp->console_in = console_in;
        readp->console_out = console_out;
    }
    return readp;
}

/*
 * Copy the current line from the history.
 */
static void LoadFromHistory(ReadLine *readp)
{
    LineBuffer *linep, *current;

    current = &readp->current;
    linep = readp->history + ((readp->start + readp->index) % MAX_HISTORY_LENGTH);
    *current = *linep;
}

/*
 * Set the console stats.
 */
static int ConsoleRaw(ReadLine *readp)
{
    CONSOLE_SCREEN_BUFFER_INFO info;
    int status;

    /* Set in raw mode */
    status = SetConsoleMode(readp->console_in, RAW_CONSOLE_MODE);
    if(status == 0) {
        print_error("SetConsoleMode");
        return -1;
    }
    status = SetConsoleMode(readp->console_out, ENABLE_PROCESSED_OUTPUT | ENABLE_WRAP_AT_EOL_OUTPUT);
    if(status == 0) {
        print_error("SetConsoleMode");
        return -1;
    }

    /* Remember where the cursor currently is */
    status = GetConsoleScreenBufferInfo(readp->console_out, &info);
    if(status == 0) {
        print_error("GetConsoleScreenBufferInfo");
        readp->x = 0;
        readp->y = 0;
        readp->w = 80;
        readp->h = 24;
    }
    else {
        readp->x = info.dwCursorPosition.X;
        readp->y = info.dwCursorPosition.Y;
        readp->w = info.dwSize.X;
        readp->h = info.dwSize.Y;
        if(readp->w <= 0)
            readp->w = 1;
        if(readp->h <= 0)
            readp->h = 1;
    }
    return 0;
}

/*
 * Set the console stats.
 */
static int ConsoleCooked(ReadLine *readp)
{
    int status;

    status = SetConsoleMode(readp->console_in, COOKED_CONSOLE_MODE);
    if(status == 0) {
        print_error("SetConsoleMode");
        return -1;
    }
    return 0;
}

/*
 * Move the cursor to the right index.
 */
static void UpdateCursor(ReadLine *readp, int index)
{
    int i, x, y;
    char *bufp;
    COORD pos;

    bufp = readp->current.buffer;
    x = readp->x;
    y = readp->y;
    for(i = 0; i != index; i++) {
        switch(bufp[i]) {
        case 0:
            break;
        case '\r':
            x = 0;
            break;
        case '\n':
            y++;
            break;
        case '\t':
            x = (x + 8) & ~7;
            break;
        default:
            x++;
            break;
        }
        if(x == readp->w) {
            x = 0;
            y++;
        }
        if(y == readp->h) {
            readp->y--;
            y = readp->h - 1;
        }
    }

    /* Set the cursor */
    pos.X = x;
    pos.Y = y;
    SetConsoleCursorPosition(readp->console_out, pos);
}

/*
 * Go to the next line in the history.
 */
static void ReadNextLine(ReadLine *readp)
{
    if(readp->index < readp->length)
        readp->index++;
    LoadFromHistory(readp);
}

static void ReadPrevLine(ReadLine *readp)
{
    if(readp->index > 0)
        readp->index--;
    LoadFromHistory(readp);
}

/*
 * Clear the screen.
 */
static void ReadClearScreen(ReadLine *readp)
{
    COORD pos;
    DWORD count;

    /* Flood the screen with whitespace */
    pos.X = 0;
    pos.Y = 0;
    count = readp->w * readp->h;
    FillConsoleOutputCharacter(readp->console_out, (TCHAR) ' ', count, pos, &count);

    /* Move the cursor and print the prompt */
    readp->x = 0;
    readp->y = 0;
    UpdateCursor(readp, 0);
    WriteFile(readp->console_out,
              readp->prompt,
              readp->prompt_length,
              &count,
              NULL);

    /* Move the standard cursor */
    readp->y = readp->prompt_length / readp->w;
    readp->x = readp->prompt_length % readp->w;
    UpdateCursor(readp, 0);
}

/*
 * Refresh the current line in the console.
 */
static void Refresh(ReadLine *readp)
{
    LineBuffer *old, *current;
    DWORD count;
    int end;

    old = &readp->display;
    current = &readp->current;

    /* Write to max of old and new lengths */
    if(current->length < old->length)
        end = old->length;
    else
        end = current->length;

    /* Copy the entire line, and pad with spaces */
    memcpy(old->buffer, current->buffer, current->length);
    memset(old->buffer + current->length, ' ', end - current->length);

    /* Write the data to the console */
    UpdateCursor(readp, 0);
    WriteFile(readp->console_out,               // Echo to the console
              old->buffer,                      // Data to write
              end,                              // Write all characters
              &count,                           // Number of bytes written
              NULL);                            // Not overlapped
    UpdateCursor(readp, current->index);
    old->index = current->index;
    old->length = current->length;
}

/************************************************************************
 * Processor.
 */

/*
 * Send the character to the current process.
 */
static ProcessCode process_char(ReadLine *readp, char c)
{
    LineBuffer *linep;
    ProcessCode code;

    code = CODE_SUCCESS;
    linep = &readp->current;
    if(readp->escape) {
        readp->escape = 0;
        LineInsertChar(linep, c);
    }
    else {
        switch(c) {
        case 0:
            break;
        case 'B'-'@':
            LineMovePrevChar(linep);
            break;
        case 'D'-'@':
            code = CODE_EOF;
            break;
        case 'F'-'@':
            LineMoveNextChar(linep);
            break;
        case 'H'-'@':
            LineDeletePrevChar(linep);
            break;
        case 'K'-'@':
            LineDeleteToEOL(linep);
            break;
        case 'L'-'@':
            ReadClearScreen(readp);
            break;
        case 'P'-'@':
            ReadPrevLine(readp);
            break;
        case 'N'-'@':
            ReadNextLine(readp);
            break;
        case 'V'-'@':
            readp->escape = 1;
            break;
        case 'W'-'@':
            LineDeletePrevWord(linep);
            break;
        case 'X'-'@':
        case 'C'-'@':
            LineDeleteLine(linep);
            break;
        case 127:
            LineDeleteNextChar(linep);
            break;
        case '\n':
        case '\r':
            LineInsertEOL(linep);
            code = CODE_EOL;
            break;
        case '\\'-'@':
        case 'Z'-'@':
            break;
        default:
            LineInsertChar(linep, c);
            break;
        }
    }
    Refresh(readp);
    return code;
}

/*
 * Input loop.
 */
static ProcessCode processor(ReadLine *readp)
{
    INPUT_RECORD input[INPUT_COUNT];
    INPUT_RECORD *event;
    KEY_EVENT_RECORD *key;
    ProcessCode code;
    int i, status;
    DWORD count;
    char c;

    /* Input loop */
    while(1) {
        enter_blocking_section();
        status = ReadConsoleInput(readp->console_in, input, INPUT_COUNT, &count);
        leave_blocking_section();
        if(status == 0) {
            print_error("ReadConsoleInput");
            return CODE_EOF;
        }

        /* Perform each event */
        for(i = 0; i != count; i++) {
            event = &input[i];
            switch(event->EventType) {
            case KEY_EVENT:
                /* Send the key to the process */
                key = &event->Event.KeyEvent;
                c = key->uChar.AsciiChar;
                if(key->bKeyDown) {
                    code = process_char(readp, c);
                    if(code != CODE_SUCCESS)
                        return code;
                }
                break;
            case MOUSE_EVENT:
            case WINDOW_BUFFER_SIZE_EVENT:
            case MENU_EVENT:
            case FOCUS_EVENT:
            default:
                break;
            }
        }
    }
}

/*
 * Read in cooked mode.
 */
static void readline_cooked(ReadLine *readp)
{
    char *s;

    enter_blocking_section();
    s = fgets(readp->current.buffer, MAX_LINE_LENGTH, stdin);
    leave_blocking_section();
    if(s == 0)
        raise_end_of_file();
    readp->current.length = strlen(readp->current.buffer);
}

/*
 * Read in raw console mode.
 */
static void readline_raw(ReadLine *readp)
{
    LineBuffer *linep;

    /* Process in raw mode */
    ConsoleRaw(readp);
    processor(readp);
    ConsoleCooked(readp);
}

/*
 * Once reading is done, reset all the buffers.
 */
static readline_done(ReadLine *readp)
{
    LineBuffer *linep;

    /* Copy the current line to the history */
    linep = readp->history + ((readp->start + readp->length) % MAX_HISTORY_LENGTH);
    *linep = readp->current;

    /* Now scroll the history */
    if(readp->length < MAX_HISTORY_LENGTH - 1)
        readp->length++;
    else
        readp->start = (readp->start + 1) % MAX_HISTORY_LENGTH;
    readp->index = readp->length;

    /* Clear the next line in the history */
    linep = readp->history + ((readp->start + readp->length) % MAX_HISTORY_LENGTH);
    linep->index = 0;
    linep->length = 0;

    /* Clear the display buffer */
    linep = &readp->display;
    linep->index = 0;
    linep->length = 0;

    /* Data is copied from the current buffer */
    linep = &readp->current;
    linep->index = 0;
}

/*
 * Read a line from the console.
 */
static void readline(ReadLine *readp, const char *promptp)
{
    LineBuffer *linep;
    int c, i;

    /* Get the prompt */
    for(i = 0; i < sizeof(readp->prompt) - 1; i++) {
        c = *promptp++;
        if(c == 0)
            break;
        readp->prompt[i] = c;
    }
    readp->prompt[i + 1] = 0;
    readp->prompt_length = i;

    /* Display the prompt */
    if(readp->is_console || readp->force_prompt) {
        printf("%s", readp->prompt);
        fflush(stdout);
    }

    /* Read the input line */
    if(readp->is_console)
        readline_raw(readp);
    else
        readline_cooked(readp);

    /* Add this line to the history */
    readline_done(readp);
}

/************************************************************************
 * Main routines.
 */

/*
 * Only one readline buffer.
 */
static ReadLine *readp;

/*
 * Is input a tty?
 */
value omake_isatty(value v_unit)
{
    return Val_int(readp->is_console);
}

value omake_is_interactive(value v_unit)
{
    return Val_int(readp->is_console || readp->force_prompt);
}

/*
 * Force the prompt anyway.
 */
value omake_interactive(value v_bool)
{
    int interactive = Int_val(v_bool);
    if(interactive)
        readp->force_prompt = 1;
    else {
        readp->is_console = 0;
        readp->force_prompt = 0;
    }
    return Val_unit;
}

/*
 * Read an entire line from the input.
 */
value omake_readline(value v_prompt)
{
    LineBuffer *linep;
    value v_str;

    /* Get the line */
    readline(readp, String_val(v_prompt));

    /* Copy it to a string */
    linep = &readp->current;
    v_str = alloc_string(linep->length);
    memcpy(String_val(v_str), linep->buffer, linep->length);

    /* Reset the current buffer */
    linep->index = 0;
    linep->length = 0;
    return v_str;
}

/*
 * Read a single character from the input.
 */
value omake_readstring(value v_prompt, value v_buf, value v_off, value v_len)
{
    LineBuffer *linep;
    int off, len, amount;
    char *buf, *prompt;

    /* If the buffer is empty, read the next line */
    prompt = String_val(v_prompt);
    linep = &readp->current;
    if(linep->index == linep->length) {
        linep->index = 0;
        linep->length = 0;
        readline(readp, prompt);
    }

    /* Get as much as possible */
    buf = String_val(v_buf);
    off = Int_val(v_off);
    len = Int_val(v_len);
    amount = linep->length - linep->index;
    if(amount > len)
        amount = len;
    memcpy(buf + off, linep->buffer + linep->index, amount);
    linep->index += amount;
    return Val_int(amount);
}

/*
 * Attach to the console.
 */
value omake_readline_init(value v_unit)
{
    HANDLE c_stdin, c_stdout, handle;
    DWORD status, mode;
    int is_console;

    /* Get the stdin handle */
    c_stdin = GetStdHandle(STD_INPUT_HANDLE);
    c_stdout = GetStdHandle(STD_OUTPUT_HANDLE);
    if(c_stdin == INVALID_HANDLE_VALUE || c_stdout == INVALID_HANDLE_VALUE)
        failwith("omake_readline_init: no standard channels");

    /* Check if it is a console */
    is_console = 1;
    status = GetConsoleMode(c_stdin, &mode);
    if(status) {
        /* Make sure output is to the console */
        handle = CreateFile("CONOUT$",
                            GENERIC_READ | GENERIC_WRITE,
                            FILE_SHARE_READ | FILE_SHARE_WRITE,
                            NULL,
                            OPEN_EXISTING,
                            FILE_ATTRIBUTE_NORMAL,
                            NULL);
        if(handle == INVALID_HANDLE_VALUE)
            is_console = 0;
        else
            c_stdout = handle;
    }
    else
        is_console = 0;

    /* Good, we have a console */
    readp = AllocReadLine(is_console, c_stdin, c_stdout);
    return Val_unit;
}

#else /* WIN32 */

#include <unistd.h>
#include <string.h>

#ifdef READLINE_ENABLED
#include <readline/readline.h>
#include <readline/history.h>
#endif

/*
 * Input state.
 */
typedef struct _readline {
    /* Console info */
    int console_in;
    int is_console;

    /* Prompts */
    char prompt[MAX_PROMPT_LENGTH];
    int force_prompt;

    /* Buffered data */
    char buffer[MAX_LINE_LENGTH];
    int index;
    int length;
} ReadLine;

/*
 * Allocate the struct.
 */
static ReadLine *AllocReadLine(int is_console, int console_in)
{
    ReadLine *readp;

    /* Allocate */
    readp = (ReadLine *) malloc(sizeof(ReadLine));
    if(readp == 0)
        failwith("AllocReadLine: out of memory");
    memset(readp, 0, sizeof(ReadLine));

    /* Initialize */
    readp->console_in = console_in;
    readp->is_console = is_console;
    return readp;
}

/*
 * Read in cooked mode.
 */
static void readline_cooked(ReadLine *readp)
{
    if(readp->is_console || readp->force_prompt) {
        printf("%s", readp->prompt);
        fflush(stdout);
    }
    if(fgets(readp->buffer, MAX_LINE_LENGTH, stdin) == NULL)
        raise_end_of_file();
    readp->length = strlen(readp->buffer);
}

/*
 * Read in raw console mode.
 */
#ifdef READLINE_ENABLED
static void readline_raw(ReadLine *readp)
{
    char *linep;
    int length;

    rl_on_new_line();
    linep = readline(readp->prompt);
    if(linep) {
        length = strlen(linep);
        if(length > MAX_LINE_LENGTH)
            length = MAX_LINE_LENGTH;
        memcpy(readp->buffer, linep, length);
        readp->buffer[length] = '\n';
        readp->length = length + 1;
        if(*linep)
            add_history(linep);
        free(linep);
    }
    else
        readp->length = 0;
}
#else /* READLINE_ENABLED */
#define readline_raw readline_cooked
#endif /* !READLINE_ENABLED */

/*
 * Read a line from the console.
 */
static void do_readline(ReadLine *readp, const char *promptp)
{
    int c, i;

    /* Get the prompt */
    for(i = 0; i < sizeof(readp->prompt) - 1; i++) {
        c = *promptp++;
        if(c == 0)
            break;
        readp->prompt[i] = c;
    }
    readp->prompt[i] = 0;

    /* Read the input line */
    if(readp->is_console)
        readline_raw(readp);
    else
        readline_cooked(readp);
    readp->index = 0;
}

/************************************************************************
 * Public functions.
 */

static ReadLine *readp;

/*
 * Is input a tty?
 */
value omake_isatty(value v_unit)
{
    return Val_int(readp->is_console);
}

value omake_is_interactive(value v_unit)
{
    return Val_int(readp->is_console || readp->force_prompt);
}

/*
 * Force the prompt anyway.
 */
value omake_interactive(value v_bool)
{
    int interactive = Int_val(v_bool);
    if(interactive)
        readp->force_prompt = 1;
    else {
        readp->is_console = 0;
        readp->force_prompt = 0;
    }
    return Val_unit;
}

/*
 * Read an entire line from the input.
 */
value omake_readline(value v_prompt)
{
    value v_str;

    /* Get the line */
    do_readline(readp, String_val(v_prompt));

    /* Copy it to the buffer */
    v_str = alloc_string(readp->length);
    memcpy(String_val(v_str), readp->buffer, readp->length);

    /* Reset the current buffer */
    readp->index = 0;
    readp->length = 0;
    return v_str;
}

/*
 * Read a single character from the input.
 */
value omake_readstring(value v_prompt, value v_buf, value v_off, value v_len)
{
    int off, len, amount;
    char *buf, *prompt;

    /* If the buffer is empty, read the next line */
    prompt = String_val(v_prompt);
    if(readp->index == readp->length) {
        readp->index = 0;
        readp->length = 0;
        do_readline(readp, prompt);
    }

    /* Get as much as possible */
    buf = String_val(v_buf);
    off = Int_val(v_off);
    len = Int_val(v_len);
    amount = readp->length - readp->index;
    if(amount > len)
        amount = len;
    memcpy(buf + off, readp->buffer + readp->index, amount);
    readp->index += amount;
    return Val_int(amount);
}

/*
 * Attach to the console.
 */
value omake_readline_init(value v_unit)
{
    readp = AllocReadLine(isatty(0), 0);
    return Val_unit;
}

#endif /* !WIN32 */
